use strict;

# First read arguments and choose library and output directory

my $precision = '';
my $color = '';
my $making_header_file = 0;
my $lib = '';
my $pc = '';
my $outdir = '';

sub get_arg($\$) {
  my($r, $i) = @_;
  if(!$r) {
    if(!($ARGV[$$i+1]=~/^-/)) {
      $$i++;
      $r = $ARGV[$$i];
    }
  }
  return $r;
}

for(my $i=0; $i<=$#ARGV; $i++) {
  for($ARGV[$i]) {
    /^-h$/ && do { $making_header_file = 1; next; };
    /^-s$/ && do { $making_header_file = 0; next; };
    /^-p(.*)/ && do { $precision = get_arg($1, $i); next; };
    /^-c(.*)/ && do { $color = get_arg($1, $i); next; };
    /^-l(.*)/ && do {
      $lib = get_arg($1, $i);
      if($lib=~/([DFdf]*)([23Nn]*)/) {
	$precision = $1;
	$color = $2;
      }
      next;
    };
    /^-d(.*)/ && do { $outdir = get_arg($1, $i); next; };
    /.*/ && do {
      print "bad argument ", $_, "\n";
      print $0, " [-h] [-s] [-p<precision>] [-c<color>] [-l<library>] [-d<dir>]\n";
      exit(1);
    };
  }
}

$precision = uc $precision;
$color = uc $color;

if( ! ($precision =~ /^$|^F$|^D$|^FD$|^DF$/) ) {
  print "bad precision ", $precision, "\n";
  exit(1);
}

if( ! ($color =~ /^$|^2$|^3$|^N$/) ) {
  print "bad color ", $color, "\n";
  exit(1);
}

$pc = $precision.$color;
if($pc) { $pc = "_".$pc; }
$lib = $pc;
if(!$lib) { $lib = '_int'; }

$outdir =~ s/(.*)[\/]*/$1\//;

my $hlib = lc $lib;
if($precision eq "FD") { $hlib =~ s/fd/df/; }
if($making_header_file) {
  open HFILE, ">>".$outdir."qdp".$hlib.".h";
  END {
    print HFILE "\n#endif\n";
    close HFILE;
  }
}

# Now define some variables

my(%datatypes);
my(@all_types, @shift_types, @float_types, @complex_types, @color_types);
scalar eval `cat datatypes.pl`;

my @eqops = ( "eq", "peq", "meq", "eqm" );

# define some subroutines

sub check_type($) {
  my($t) = @_;
  if($t) {
    if(!$datatypes{$t}{ABBR}) {
      print "$0: error bad datatype $t\n"; 
      exit 1;
    }
  }
}

sub correct_precision_and_color($$$) {
  my($t1, $t2, $t3) = @_;
  check_type($t1);
  check_type($t2);
  check_type($t3);
  my($has_precision) = 0;
  if( (($t1)&&(!$datatypes{$t1}{NO_PRECISION})) ||
      (($t2)&&(!$datatypes{$t2}{NO_PRECISION})) ||
      (($t3)&&(!$datatypes{$t3}{NO_PRECISION})) ) { $has_precision = 1; };
  my($has_color) = 0;
  if( (($t1)&&(!$datatypes{$t1}{NO_COLOR})) ||
      (($t2)&&(!$datatypes{$t2}{NO_COLOR})) ||
      (($t3)&&(!$datatypes{$t3}{NO_COLOR})) ) { $has_color = 1; };
  my($correct) = 1;
  if($precision) {
    if(!$has_precision) { $correct = 0; };
  } else {
    if($has_precision) { $correct = 0; };
  }
  if($color) {
    if(!$has_color) { $correct = 0; };
  } else {
    if($has_color) { $correct = 0; };
  }
  return $correct;
}

sub is_scalar($) {
  my($abbr) = @_;
  return $abbr =~ /^_[a-z]/;
}

sub do_subset($$$$$$$$$$) {
  my($indent, $qla1, $qla2, $y0, $dest, $y1, $src1, $y2, $src2, $y3) = @_;
  my($v) = 'v';
  my($x) = 'x';
  my($dv,$dvi);
  my($multi1, $multi2);
  my($subset) = "subset";
  if($dest->{SCALAR}) {
    if($dest->{EXTENDED}) {
      if($dest->{MULTI}) {
	$dv = "&dtemp[i]";
	$subset = "subset[i]";
      } else {
	$dv = "&dtemp";
      }
      $dvi = $dv;
    } else {
      if($dest->{MULTI}) {
	$dv = "&dest[i]";
	$subset = "subset[i]";
      } else {
	$dv = "dest";
      }
      $dvi = $dv;
    }
  } else {
    $dv  = "dest->data+$subset->offset";
    $dvi = "dest->data";
  }
  my($s1,$s1i);
  $s1i = $src1;
  $s1 = $src1."+$subset->offset" if($s1i);
  my($s2,$s2i);
  $s2i = $src2;
  $s2 = $src2."+$subset->offset" if($s2i);
  my($args) = "$y0 $dv$y1$s1$y2$s2$y3";
  my($argsi) = "$y0 $dvi$y1$s1i$y2$s2i$y3";
  my($body) = <<"  EOF;";
  if($subset->indexed==0) {
    $qla1$v$qla2($args, $subset->len );
  } else {
    $qla1$x$qla2($args, $subset->index, $subset->len );
  }
  EOF;
  my($space);
  if($dest->{MULTI}) {
    $body = "int i;\nfor(i=0; i<ns; i++) {\n".$body;
    $body .= "}\n";
    $space = ' ' x ($indent);
  } else {
    $space = ' ' x ($indent-2);
  }
  $body =~ s/^/$space/gm;
  return $body;
}

sub qla_ext_type($) {
  my($dt) = @_;
  my($type) = $dt->{TYPE};
  my($tpc) = $dt->{PC};
  $tpc =~ s/^_F/_D/ ||
  $tpc =~ s/^_D/_Q/;
  my($name);
  $name = "QLA".$tpc."_".$type;
  return $name;
}

sub prepare_dest($) {
  my($dest) = @_;
  my($abbr) = $dest->{ABBR};
  my($tpc) = $dest->{PC};
  my($r) = '';
  if($dest->{SCALAR}) {
    if($dest->{EXTENDED}) {
      my($ext) = qla_ext_type($dest);
      if($dest->{MULTI}) {
	$r = "  ".$ext." *dtemp;\n";
	$r .= "  dtemp = ($ext *) malloc(ns*sizeof($ext));\n";
      } else {
	$r = "  ".$ext." dtemp;\n";
      }
    }
  } else {
    $r = "  QDP_prepare_dest(&dest->dc);\n";
  }
  return $r;
}

sub prepare_src($$) {
  my($src, $v) = @_;
  my($abbr) = $src->{ABBR};
  my($tpc) = $src->{PC};
  my($r) = '';
  if($abbr=~/^_[A-Z]/) { $r = "  QDP_prepare_src(&$v->dc);\n"; }
  return $r;
}

sub global_sum($) {
  my($dest) = @_;
  my($r);
  my($abbr) = $dest->{ABBR};
  my($uabbr) = uc $abbr;
  my($tpc) = $dest->{PC};
  my($epc,$cpc) = {'',''};
  if($tpc =~ /^_F/) {
    ($epc = $tpc) =~ s/^_F/_D/;
    ($cpc = $tpc) =~ s/^_F/_FD/;
  } else {
    ($epc = $tpc) =~ s/^_D/_Q/;
    ($cpc = $tpc) =~ s/^_D/_DQ/;
  }
  my($nc,$ncn) = ('','');
  if(($color eq 'N')&&(!$datatypes{$dest->{TYPE}}{NO_COLOR})) {
    $nc = 'nc, ';
    $ncn = '_N';
  }
  if($dest->{MULTI}) {
    if($dest->{EXTENDED}) {
      $r = "  QDP".$ncn."_binary_reduce_multi(".$nc."QLA".$epc.$uabbr."_vpeq".$uabbr.", sizeof(QLA".$epc."_".$dest->{TYPE}."), dtemp, ns);\n";
      $r .= "  QLA".$cpc.$uabbr."_veq".$uabbr."(".$nc."dest, dtemp, ns);\n";
      $r .= "  free(dtemp);\n"
    } else {
      $r = "  QDP".$ncn."_binary_reduce_multi(".$nc."QLA".$tpc.$uabbr."_vpeq".$uabbr.", sizeof(QLA".$tpc."_".$dest->{TYPE}."), dest, ns);\n";
    }
  } else {
    if($dest->{EXTENDED}) {
      $r = "  QDP".$ncn."_binary_reduce(".$nc."QLA".$epc.$uabbr."_peq".$uabbr.", sizeof(QLA".$epc."_".$dest->{TYPE}."), &dtemp);\n";
      $r .= "  QLA".$cpc.$uabbr."_eq".$uabbr."(".$nc."dest, &dtemp);\n";
    } else {
      $r = "  QDP".$ncn."_binary_reduce(".$nc."QLA".$tpc.$uabbr."_peq".$uabbr.", sizeof(QLA".$tpc."_".$dest->{TYPE}."), dest);\n";
    }
  }

  if (0) {
  for($dest->{TYPE}) {
    /Int/ && do {
      if($dest->{MULTI}) {
	$r =  "  {\n";
	$r .= "    int i;\n";
	$r .= "    for(i=0; i<ns; i++) {\n";
	$r .= "      g_intsum( &dest[i] );\n";
	$r .= "    }\n";
	$r .= "  }\n";
      } else {
	$r = "  g_intsum( dest );\n";
      }
      last;
    };
    /Real/ && do {
      my($vec, $ns);
      if($dest->{MULTI}) {
	$vec = "vec";
	$ns = ", ns";
      }
      if($precision eq 'F') {
	$r = "  g_".$vec."floatsum( dest".$ns." );\n";
      } else {
	$r = "  g_".$vec."doublesum( dest".$ns." );\n";
      }
      last;
    };
    /Complex/ && do {
      my($vec, $ns);
      if($dest->{MULTI}) {
	$vec = "vec";
	$ns = ", ns";
      }
      if($precision eq 'F') {
	$r = "  g_".$vec."complexsum( (complex *) dest".$ns." );\n";
      } else {
	$r = "  g_".$vec."dcomplexsum( (double_complex *) dest".$ns." );\n";
      }
      last;
    };
    /.*/ && do {
      my($ns);
      if($dest->{MULTI}) {
	$ns = " * ns";
      }
      if($precision eq 'F') {
	$r = "  g_veccomplexsum( (complex *) dest, $datatypes{$_}{CSIZE}".$ns." );\n";
      } else {
	$r = "  g_vecdcomplexsum( (double_complex *) dest, $datatypes{$_}{CSIZE}".$ns." );\n";
      }
      last;
    };
  }
  }
  return $r;
}

sub body0($$$$$) {
  my($qla1, $qla2, $y0, $dest, $y1) = @_;
  my($body) = "{\n";
  $body .= prepare_dest($dest);
  $body .= "\n";
  $body .= do_subset(2, $qla1, $qla2, $y0, $dest, $y1, '', '', '', '');
  $body .= "}\n";
  return $body;
}

sub body1ns($$$$$$$$$) {
  my($qla1, $qla2, $qla3, $y0, $dest, $y1, $src, $sv, $y2) = @_;
  my($ds) = $dest->{SCALAR};
  my($body) = "{\n";
  $body .= prepare_dest($dest);
  $body .= prepare_src($src, $sv);
  $body .= "\n";
  $body .= do_subset(2, $qla1, $qla2.$qla3, $y0, $dest, $y1, ", src->data", $y2, '', '');
  $body .= "}\n";
  return $body;
}

sub body1($$$$$$$$$) {
  my($qla1, $qla2, $qla3, $y0, $dest, $y1, $src, $sv, $y2) = @_;
  my($body) = "{\n";
  $body .= prepare_dest($dest);
  $body .= prepare_src($src, $sv);
  $body .= "\n";
  $body .= "  if($sv->ptr==NULL) {\n";
  $body .= do_subset(4, $qla1, $qla2.$qla3, $y0, $dest, $y1, ", $sv->data", $y2, '', '');
  $body .= "  } else {\n";
  $body .= do_subset(4, $qla1, $qla2.'p'.$qla3, $y0, $dest, $y1, ", $sv->ptr", $y2, '', '');
  $body .= "  }\n";
  if($dest->{SCALAR}) { $body .= global_sum($dest); }
  $body .= "}\n";
  return $body;
}

sub body2($$$$$$$$$$$$$) {
  my($qla1, $qla2, $qla3, $qla4, $y0, $dest, $y1, $src1, $s1v, $y2, $src2, $s2v, $y3) = @_;
  my($ds) = $dest->{SCALAR};
  my($body) = "{\n";
  $body .= prepare_dest($dest);
  $body .= prepare_src($src1, $s1v);
  $body .= prepare_src($src2, $s2v);
  $body .= "\n";
  $body .= "  if(src1->ptr==NULL) {\n";
  $body .= "    if(src2->ptr==NULL) {\n";
  $body .= do_subset(6, $qla1, $qla2.$qla3.$qla4, $y0, $dest, $y1, ", src1->data", $y2, ", src2->data", $y3);
  $body .= "    } else {\n";
  $body .= do_subset(6, $qla1, $qla2.$qla3.'p'.$qla4, $y0, $dest, $y1, ", src1->data", $y2, ", src2->ptr", $y3);
  $body .= "    }\n";
  $body .= "  } else {\n";
  $body .= "    if(src2->ptr==NULL) {\n";
  $body .= do_subset(6, $qla1, $qla2.'p'.$qla3.$qla4, $y0, $dest, $y1, ", src1->ptr", $y2, ", src2->data", $y3);
  $body .= "    } else {\n";
  $body .= do_subset(6, $qla1, $qla2.'p'.$qla3.'p'.$qla4, $y0, $dest, $y1, ", src1->ptr", $y2, ", src2->ptr", $y3);
  $body .= "    }\n";
  $body .= "  }\n";
  if($ds) { $body .= global_sum($dest); }
  $body .= "}\n";
  return $body;
}

sub qla_name($$$$$) {
  my($dest, $op, $s1aa, $func, $s2aa) = @_;
  my($tpc) = $pc;
  if($dest->{EXTENDED}) {
    $tpc =~ s/^_F/_DF/ ||
      $tpc =~ s/^_D/_QD/;
  }
  my($qla1, $qla2, $qla3, $qla4);
  $qla1 = "QLA".$tpc.$dest->{ABBR}."_";
  $qla2 = $op."_";
  $qla3 = $s1aa.$func;
  $qla3 =~ s/^_//;
  $qla4 = $s2aa;
  $qla4 =~ s/^_//;
  if($qla4) { $qla3 .= '_'; }
  return ($qla1, $qla2, $qla3, $qla4);
}

sub func_body($$$$$$$$) {
  my($dest, $op, $src1, $func, $src2, $x1, $x2, $x3) = @_;

  my($dt) = $dest->{TYPE};
  my($dabbr) = $dest->{ABBR};

  my($s1t) = $src1->{TYPE};
  my($s1abbr) = $src1->{ABBR};
  my($s1adj) = $src1->{ADJ};

  my($s2t) = $src2->{TYPE};
  my($s2abbr) = $src2->{ABBR};
  my($s2adj) = $src2->{ADJ};

  if( ($dest->{SCALAR}) &&
      ((!$s1t)||(!$datatypes{$s1t}{NO_PRECISION})) &&
      ((!$s2t)||(!$datatypes{$s2t}{NO_PRECISION})) ) {
    $dest->{EXTENDED} = 1;
  }

  my($qla1, $qla2, $qla3, $qla4) = qla_name($dest, $op, $s1abbr.$s1adj, $func, $s2abbr.$s2adj);

  my($y0) = ('');
  if($color eq 'N') { $y0 = ' nc,'; }
  my($y1) = $x1;
  $y1 =~ s/[^, ]*[ *]+([^, *]+),/$1,/g;
  my($y2) = $x2;
  if($func eq '_func') {
    $y2 =~ s/\([^()]*\)([^()]*)$/$1/;
    $y2 =~ s/[()]*//g;
  }
  $y2 =~ s/[^, ]*[ *]+([^, *]+),/$1,/g;
  my($y3) = $x3;
  $y3 =~ s/[^, ]*[ *]+([^, *]+),/$1,/g;
  if($y1) { $y1 = ", ".$y1; $y1 =~ s/[, ]*$//; }
  if($y2) { $y2 = ", ".$y2; $y2 =~ s/[, ]*$//; }
  if($y3) { $y3 = ", ".$y3; $y3 =~ s/[, ]*$//; }

  my($nsrc, $src, $srcvar);
  my($body);
  $nsrc = 2;
  $srcvar = "src1";
  if(!$s2t) {
    $nsrc--;
    $qla3 .= $qla4;
    $qla4 = "";
    $y2 .= $y3;
    $y3 = "";
    $src = $src1;
    $srcvar = 'src';
  }
  if(!$s1t) {
    $nsrc--;
    $qla2 .= $qla3;
    $qla3 = $qla4;
    $y1 .= $y2;
    $y2 = $y3;
    $src = $src2;
    $srcvar = 'src';
  } elsif(is_scalar($s1abbr)) {
    $nsrc--;
    $qla2 .= $qla3;
    $qla3 = $qla4;
    $y1 .= ", ".$srcvar.$y2;
    $y2 = $y3;
    $src = $src2;
    $srcvar = 'src2';
  }

  if($nsrc==0) {
    $body = body0($qla1, $qla2, $y0, $dest, $y1);
  } elsif($nsrc==1) {
    if($datatypes{$src->{TYPE}}{NO_SHIFT}) {
      $body = body1ns($qla1, $qla2, $qla3, $y0, $dest, $y1, $src, $srcvar, $y2);
    } else {
      $body = body1($qla1, $qla2, $qla3, $y0, $dest, $y1, $src, $srcvar, $y2);
    }
  } else {
    $body = body2($qla1, $qla2, $qla3, $qla4, $y0, $dest, $y1, $src1, 'src1', $y2, $src2, 'src2', $y3);
  }
  return $body;
}

sub qdp_name($$$$$) {
  my($dest, $op, $s1aa, $func, $s2aa) = @_;
  my($da) = $dest->{ABBR};
  my($multi) = "";
  if($dest->{MULTI}) { $multi = "_multi"; }
  return "QDP".$pc.$da."_".$op.$s1aa.$func.$s2aa.$multi;
}

sub type_name($) {
  my($typehash) = @_;
  my($type) = $typehash->{TYPE};
  my($scalar) = $typehash->{SCALAR};
  my($p);
  if($precision eq 'FD') {
    if($typehash->{DEST}) { $p = "F"; }
    else { $p = "D"; }
  } elsif($precision eq 'DF') {
    if($typehash->{DEST}) { $p = "D"; }
    else { $p = "F"; }
  } else {
    $p = $precision;
  }
  my($tpc);
  if($datatypes{$type}{NO_PRECISION}) {
    $tpc = '';
  } elsif($datatypes{$type}{NO_COLOR}) {
    $tpc = '_'.$p;
  } else {
    $tpc = '_'.$p.$color;
  }
  $typehash->{PC} = $tpc;
  my($name);
  if($scalar) {
    $name = "QLA".$tpc."_".$type;
  } else {
    $name = "QDP".$tpc."_".$type;
  }
  return $name;
}

sub qdp_args($$$$$$) {
  my($dest, $x1, $src1, $x2, $src2, $x3) = @_;
  my($s1t) = $src1->{TYPE};
  my($s2t) = $src2->{TYPE};
  if($s1t) {
    if($s2t) {
      $src1->{VAR} = 'src1';
      $src2->{VAR} = 'src2';
    } else {
      $src1->{VAR} = 'src';
      $src2->{VAR} = '';
    }
  } else {
    $src1->{VAR} = '';
    if($s2t) {
      $src2->{VAR} = 'src';
    } else {
      $src2->{VAR} = '';
    }
  }
  my($s1arg) = '';
  if($s1t) { $s1arg = type_name($src1)." *".$src1->{VAR}.", "; }
  my($s2arg) = '';
  if($s2t) { $s2arg = type_name($src2)." *".$src2->{VAR}.", "; }
  my($args) = "( ";
  if($color eq 'N') { $args .= 'int nc, '; }
  $args .= type_name($dest);
  if($dest->{MULTI}) {
    $args .= " dest[], ";
  } else {
    $args .= " *dest, ";
  }
  $args .= $x1.$s1arg.$x2.$s2arg.$x3;
  if($dest->{MULTI}) {
    $args .= "QDP_Subset subset[], int ns )";
  } else {
    $args .= "QDP_Subset subset )";
  }
  return $args;
}

my($comment0, $comment1);

sub comment0($) {
  ($comment0) = @_;
}

sub comment1($) {
  ($comment1) = @_;
}

sub get_h_comment() {
  my($c) = '';
  if($comment0) {
    $c .= "\n";
    $c .= "/*\n";
    $c .= " *  ".$comment0."\n";
    $c .= " */\n";
    $c .= "\n" if(!$comment1);
    $comment0 = "";
  }
  if($comment1) {
    $c .= "\n";
    $c .= "/*\n";
    $c .= " *  ".$comment1."\n";
    $c .= " */\n";
    $c .= "\n";
    $comment1 = "";
  }
  return $c;
}

sub get_c_comment() {
  my($c) = '';
  $c .= "/*\n";
  if($comment0) {
    $c .= " *  ".$comment0."\n";
    $c .= " *\n" if($comment1);
  }
  if($comment1) {
    $c .= " *  ".$comment1."\n";
  }
  $c .= " */\n";
  $c .= "\n";
  return $c;
}

sub write_function(\%$\%$\%\%) {
  my($dest, $op, $src1, $func, $src2, $xargs) = @_;

  my($dt) = $dest->{TYPE};
  my($dscalar) = $dest->{SCALAR};

  my($s1t) = $src1->{TYPE};
  my($s1scalar) = $src1->{SCALAR};
  my($s1adj) = $src1->{ADJ};
  #$s1adj = 'a' if $src1->{ADJ};

  my($s2t) = $src2->{TYPE};
  my($s2scalar) = $src2->{SCALAR};
  my($s2adj) = $src2->{ADJ};
  #$s2adj = 'a' if $src2->{ADJ};

  if( correct_precision_and_color($dt, $s1t, $s2t) ) {

    my($dabbr, $s1abbr, $s2abbr) = ('','','');
    if($dt) {
      $dabbr = "_".$datatypes{$dt}{ABBR};
      if($dscalar) { $dabbr = lc $dabbr; }
    }
    if($s1t) {
      $s1abbr = "_".$datatypes{$s1t}{ABBR};
      if($s1scalar) { $s1abbr = lc $s1abbr; }
    }
    if($s2t) {
      $s2abbr = $datatypes{$s2t}{ABBR};
      if($func ne '__plus_i') { $s2abbr = '_'.$s2abbr; }
      if($s2scalar) { $s2abbr = lc $s2abbr; }
    }
    if($func) { $func = "_".$func; }

    $dest->{ABBR} = $dabbr;
    $src1->{ABBR} = $s1abbr;
    $src2->{ABBR} = $s2abbr;

    my($qladest) = type_name($dest);
    $qladest =~ s/QDP/QLA/;
    my($x1) = $xargs->{PROT1};
    $x1 =~ s/QLAPCTYPE/$qladest/;
    my($x2) = $xargs->{PROT2};
    $x2 =~ s/QLAPCTYPE/$qladest/;
    my($x3) = $xargs->{PROT3};
    $x3 =~ s/QLAPCTYPE/$qladest/;

    my($name) = qdp_name($dest, $op, $s1abbr.$s1adj, $func, $s2abbr.$s2adj);
    my($args) = qdp_args($dest, $x1, $src1, $x2, $src2, $x3);

    if($making_header_file) {
      my($out) = "void ".$name.$args.";\n";
      print HFILE get_h_comment();
      print HFILE $out;
    } else {
      my($head) = "void\n".$name.$args."\n";
      my($body) = func_body($dest, $op, $src1, $func, $src2, $x1, $x2, $x3);
      open CFILE, '>'.$outdir.$name.".c";
      print CFILE get_c_comment();
      print CFILE "#include \"qdp".$hlib."_internal.h\"\n\n";
      print CFILE $head.$body;
      close CFILE;
    }
  }
}

sub make_functions(\%) {
  my($arg) = @_;

  if($arg->{DO_FD}) {
    if($precision eq 'DF') { $precision = 'FD'; }
    if($precision ne 'FD') { return; }
  } elsif($arg->{DO_DF}) {
    if($precision eq 'FD') { $precision = 'DF'; }
    if($precision ne 'DF') { return; }
  } elsif(($precision eq 'FD')||($precision eq 'DF')) {
    return;
  }
  $pc = $precision.$color;
  if($pc) { $pc = "_".$pc; }

  my($dscalar) = $arg->{DEST_SCALAR};
  my($s1scalar) = $arg->{SRC1_SCALAR};
  my($s2scalar) = $arg->{SRC2_SCALAR};

  my(%xargs);
  $xargs{PROT1} = $arg->{EXTRA_ARGS1};
  $xargs{PROT2} = $arg->{EXTRA_ARGS2};
  $xargs{PROT3} = $arg->{EXTRA_ARGS3};

  my(@funcs);
  if($arg->{'FUNCS'}) { @funcs = @{$arg->{'FUNCS'}}; }
  if(!@funcs) { @funcs = ( '' ); }
  for my $func (@funcs) {

    for my $op (@{$arg->{'EQ_OPS'}}) {

      for my $dt (@{$arg->{'DEST_TYPES'}}) {

	my(@src1_types);
	if($arg->{'SRC1_TYPES'}) { @src1_types = @{$arg->{'SRC1_TYPES'}}; }
	if(!@src1_types) { @src1_types = ( '' ); }
	if($src1_types[0] eq 'DEST') { @src1_types = ( "$dt" ); }
	for my $s1t (@src1_types) {

	  my(@src1_adjs);
	  if( ($arg->{SRC1_DO_ADJ}) && ($s1t ne '') &&
	      (!$datatypes{$s1t}{NO_ADJ}) ) { @src1_adjs = ( '', 'a' ); }
	  elsif( ($arg->{SRC1_ADJ}) && ($s1t ne '') &&
		 (!$datatypes{$s1t}{NO_ADJ}) ) { @src1_adjs = ( 'a' ); }
	  else { @src1_adjs = ( '' ); }
	  for my $s1adj (@src1_adjs) {

	    my(@src2_types);
	    if($arg->{'SRC2_TYPES'}) { @src2_types = @{$arg->{'SRC2_TYPES'}}; }
	    if(!@src2_types) { @src2_types = ( '' ); }
	    if($src2_types[0] eq 'DEST') { @src2_types = ( "$dt" ); }
	    if($src2_types[0] eq 'SRC1') { @src2_types = ( "$s1t" ); }
	    for my $s2t (@src2_types) {

	      my(@src2_adjs);
	      if( ($arg->{SRC2_DO_ADJ}) && ($s2t ne '') &&
		  (!$datatypes{$s2t}{NO_ADJ}) ) { @src2_adjs = ( '', 'a' ); }
	      elsif( ($arg->{SRC2_ADJ}) && ($s2t ne '') &&
		     (!$datatypes{$s2t}{NO_ADJ}) ) { @src2_adjs = ( 'a' ); }
	      else { @src2_adjs = ( '' ); }
	      for my $s2adj (@src2_adjs) {

		my %dest = ( TYPE=>$dt, DEST=>1 );
		if($dscalar) { $dest{SCALAR} = 1; }
		if($arg->{DEST_MULTI}) { $dest{MULTI} = 1; }

		my %src1 = ( TYPE=>$s1t );
		if($s1scalar) { $src1{SCALAR} = 1; }
		#if($s1adj) { $src1{ADJ} = 1; }
		$src1{ADJ} = $s1adj;

		my %src2 = ( TYPE=>$s2t );
		if($s2scalar) { $src2{SCALAR} = 1; }
		#if($s2adj) { $src2{ADJ} = 1; }
		$src2{ADJ} = $s2adj;

		write_function(%dest, $op, %src1, $func, %src2, %xargs);
	      } # src2 adj
	    } # src2
	  }  # src1 adj
	}  # src1
      }  # dest
    }  # eqop
  }  # func
}

# Last we actually create the functions

comment0("Unary Operations");
comment1("R = <func>( R )");

make_functions(%{{
  DEST_TYPES => [ 'Real' ],
  EQ_OPS     => [ 'eq' ],
  FUNCS      => [ 'sin','cos','tan','asin','acos','atan','sqrt','fabs','exp','log','sign','ceil','floor','sinh','cosh','tanh','log10' ],
  SRC2_TYPES => [ 'DEST' ],
}});

comment1("C = exp( i R )");

make_functions(%{{
  DEST_TYPES => [ 'Complex' ],
  EQ_OPS     => [ 'eq' ],
  FUNCS      => [ 'cexpi' ],
  SRC2_TYPES => [ 'Real' ],
}});

comment1("R = <func>( C )");

make_functions(%{{
  (
   DEST_TYPES => [ 'Real' ],
   EQ_OPS     => [ 'eq' ],
   FUNCS      => [ 'norm','arg' ],
   SRC2_TYPES => [ 'Complex' ],
  )}});

comment1("C = <func>( C )");

make_functions(%{
  {
  (
   DEST_TYPES => [ 'Complex' ],
   EQ_OPS     => [ 'eq' ],
   FUNCS      => [ 'cexp','csqrt','clog' ],
   SRC2_TYPES => [ 'Complex' ],
  )}});

comment1("<field> <eqop> <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @shift_types ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> <eqop> adjoint(<field>)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex','ColorMatrix','DiracPropagator' ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'DEST' ],
   SRC1_ADJ    => 1,
  )}});

comment1("<field> <eqop> conjugate(<field>)");

make_functions(%{{
  (
   DEST_TYPES  => [ @complex_types ],
   EQ_OPS      => [ @eqops ],
   FUNCS       => [ 'conj' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> <eqop> transpose(<field>)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix','DiracPropagator' ],
   EQ_OPS      => [ @eqops ],
   FUNCS       => [ 'transpose' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("R = local norm2 <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'norm2' ],
   SRC2_TYPES  => [ @complex_types ],
  )}});


comment0("type conversion and component extraction and insertion");
comment1("(double)<field> = (float)<field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @float_types ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   DO_DF       => 1,
  )}});

comment1("(float)<field> = (double)<field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @float_types ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   DO_FD       => 1,
  )}});

comment1("C = R + i0");

make_functions(%{{
  (
   DEST_TYPES => [ 'Complex' ],
   EQ_OPS     => [ 'eq' ],
   SRC1_TYPES => [ 'Real' ],
  )}});

comment1("C = R + i R");

make_functions(%{{
  (
   DEST_TYPES => [ 'Complex' ],
   EQ_OPS     => [ 'eq' ],
   SRC1_TYPES => [ 'Real' ],
   FUNCS      => [ 'plus_i' ],
   SRC2_TYPES => [ 'Real' ],
  )}});

comment1("R = <re/im>( C )");

make_functions(%{{
  (
   DEST_TYPES => [ 'Real' ],
   EQ_OPS     => [ 'eq' ],
   FUNCS      => [ 're','im' ],
   SRC2_TYPES => [ 'Complex' ],
  )}});

comment1("R = I");

make_functions(%{{
  (
   DEST_TYPES => [ 'Real' ],
   EQ_OPS     => [ 'eq' ],
   SRC1_TYPES => [ 'Int' ],
  )}});

comment1("I = <round/trunc>( R )");

make_functions(%{{
  (
   DEST_TYPES => [ 'Int' ],
   EQ_OPS     => [ 'eq' ],
   FUNCS      => [ 'trunc','round' ],
   SRC2_TYPES => [ 'Real' ],
  )}});

comment1("C = M[i,j]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'ColorMatrix' ],
   EXTRA_ARGS3 => "int i, int j, ",
  )}});

comment1("M[i,j] = C");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'Complex' ],
   EXTRA_ARGS3 => "int i, int j, ",
  )}});

comment1("C = <H/D>[c,s]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'HalfFermion','DiracFermion' ],
   EXTRA_ARGS3 => "int color, int spin, ",
  )}});

comment1("<H/D>[c,s] = C");

make_functions(%{{
  (
   DEST_TYPES  => [ 'HalfFermion','DiracFermion' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'Complex' ],
   EXTRA_ARGS3 => "int color, int spin, ",
  )}});

comment1("C = V[i]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'ColorVector' ],
   EXTRA_ARGS3 => "int color, ",
  )}});

comment1("V[i] = C");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorVector' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'Complex' ],
   EXTRA_ARGS3 => "int color, ",
  )}});

comment1("C = P[ic,is,jc,js]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'DiracPropagator' ],
   EXTRA_ARGS3 => "int ic, int is, int jc, int js, ",
  )}});

comment1("P[ic,is,jc,js] = C");

make_functions(%{{
  (
   DEST_TYPES  => [ 'DiracPropagator' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'elem' ],
   SRC2_TYPES  => [ 'Complex' ],
   EXTRA_ARGS3 => "int ic, int is, int jc, int js, ",
  )}});

comment1("V[i] = M[i,j]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorVector' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'colorvec' ],
   SRC2_TYPES  => [ 'ColorMatrix' ],
   EXTRA_ARGS3 => "int j, ",
  )}});

comment1("M[i,j] = V[i]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'colorvec' ],
   SRC2_TYPES  => [ 'ColorVector' ],
   EXTRA_ARGS3 => "int j, ",
  )}});

comment1("V[i] = <H,D>[i,j]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorVector' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'colorvec' ],
   SRC2_TYPES  => [ 'HalfFermion','DiracFermion' ],
   EXTRA_ARGS3 => "int spin, ",
  )}});

comment1("<H,D>[i,j] = V[i]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'HalfFermion','DiracFermion' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'colorvec' ],
   SRC2_TYPES  => [ 'ColorVector' ],
   EXTRA_ARGS3 => "int spin, ",
  )}});

comment1("D[ic,is] = P[ic,is,jc,js]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'DiracFermion' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'diracvec' ],
   SRC2_TYPES  => [ 'DiracPropagator' ],
   EXTRA_ARGS3 => "int jc, int js, ",
  )}});

comment1("P[ic,is,jc,js] = D[ic,is]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'DiracPropagator' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'diracvec' ],
   SRC2_TYPES  => [ 'DiracFermion' ],
   EXTRA_ARGS3 => "int jc, int js, ",
  )}});

comment1("C = Trace(M)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'trace' ],
   SRC2_TYPES  => [ 'ColorMatrix' ],
  )}});

comment1("R = <Re/Im> Trace(M)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 're_trace','im_trace' ],
   SRC2_TYPES  => [ 'ColorMatrix' ],
  )}});

comment1("M = Anti-Hermitian Part(M)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'antiherm' ],
   SRC2_TYPES  => [ 'ColorMatrix' ],
  )}});

comment1("M = Spin Trace(P)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'spintrace' ],
   SRC2_TYPES  => [ 'DiracPropagator' ],
  )}});

comment1("H = spin project(D,dir,sign)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'HalfFermion' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'spproj' ],
   SRC2_TYPES  => [ 'DiracFermion' ],
   EXTRA_ARGS3 => "int dir, int sign, ",
  )}});

comment1("D = spin reconstruct(H,dir,sign)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'DiracFermion' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'sprecon' ],
   SRC2_TYPES  => [ 'HalfFermion' ],
   EXTRA_ARGS3 => "int dir, int sign, ",
  )}});

comment0("Binary operations with constants");

comment1("I <eqop> i times I");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Int' ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'Int' ],
   SRC1_SCALAR => 1,
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> <eqop> r times <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @float_types ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'Real' ],
   SRC1_SCALAR => 1,
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> <eqop> c times <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @complex_types ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'Complex' ],
   SRC1_SCALAR => 1,
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> <eqop> i <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @complex_types ],
   EQ_OPS      => [ @eqops ],
   FUNCS       => [ 'i' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> = gamma(dir) times <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'DiracFermion','DiracPropagator' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'gamma_times' ],
   SRC2_TYPES  => [ 'DEST' ],
   EXTRA_ARGS3 => "int dir, ",
  )}});

comment1("<field> = <field> times gamma(dir)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'DiracPropagator' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'times_gamma' ],
   EXTRA_ARGS2 => "int dir, ",
  )}});


comment0("Binary operations with fields");
comment1("<field> = <field> / <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Int','Real','Complex' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'divide' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> = <field> + <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @shift_types ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'plus' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> = <field> - <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @shift_types ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'minus' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> <eqop> <field> * <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Int','Real','Complex','DiracPropagator' ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("C = <field> local dot <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ @complex_types ],
   FUNCS       => [ 'dot' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("R = Re <field> local dot <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   EQ_OPS      => [ 'eq_re' ],
   SRC1_TYPES  => [ @complex_types ],
   FUNCS       => [ 'dot' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("M <eqop> V outer product V");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix' ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'ColorVector' ],
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'ColorVector' ],
   SRC2_ADJ    => 1,
  )}});

comment1("<field> <eqop> M times <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @color_types ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'ColorMatrix' ],
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> <eqop> adjoint(M) times <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @color_types ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'ColorMatrix' ],
   SRC1_ADJ    => 1,
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("P = P times M");

make_functions(%{{
  (
   DEST_TYPES  => [ 'DiracPropagator' ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'ColorMatrix' ],
  )}});

comment1("<M,P> = <M,P> times adjoint(M)");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix','DiracPropagator' ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'ColorMatrix' ],
   SRC2_ADJ    => 1,
  )}});

comment1("M <eqop> adjoint( M ) times adjoint( M )");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix' ],
   EQ_OPS      => [ @eqops ],
   SRC1_TYPES  => [ 'DEST' ],
   SRC1_ADJ    => 1,
   FUNCS       => [ 'times' ],
   SRC2_TYPES  => [ 'DEST' ],
   SRC2_ADJ    => 1,
  )}});

comment1("I = I <op> I");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Int' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'lshift','rshift','mod','max','min','or','and','xor' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("R = R <op> R");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'mod','max','min','pow','atan2' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("R = R *2^ I");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'ldexp' ],
   SRC2_TYPES  => [ 'Int' ],
  )}});

comment0("Ternary operations with fields");
comment1("I = i * I +/- I");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Int' ],
   EQ_OPS      => [ 'eq_i_times' ],
   EXTRA_ARGS1 => "QLA_Int *r, ",
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'plus','minus' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> = r * <field> +/- <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @float_types ],
   EQ_OPS      => [ 'eq_r_times' ],
   EXTRA_ARGS1 => "QLA_".$precision."_Real *r, ",
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'plus','minus' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("<field> = c * <field> +/- <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @complex_types ],
   EQ_OPS      => [ 'eq_c_times' ],
   EXTRA_ARGS1 => "QLA_".$precision."_Complex *c, ",
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'plus','minus' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment0("Boolean and bit operations");
comment1("comparisons");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Int' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'Int','Real' ],
   FUNCS       => [ 'eq','ne','gt','lt','ge','le' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("bitwise not");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Int' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'not' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("copy under a mask");

make_functions(%{{
  (
   DEST_TYPES  => [ @shift_types ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   FUNCS       => [ 'mask' ],
   SRC2_TYPES  => [ 'Int' ],
  )}});

comment0("Reductions");
comment1("r = norm2( <field> )");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   DEST_SCALAR => 1,
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'norm2' ],
   SRC2_TYPES  => [ @shift_types ],
  )}});

comment1("r = <field> dot <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   DEST_SCALAR => 1,
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'Int','Real' ],
   FUNCS       => [ 'dot' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("c = <field> dot <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   DEST_SCALAR => 1,
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ @complex_types ],
   FUNCS       => [ 'dot' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("r = Re <field> dot <field>");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   DEST_SCALAR => 1,
   EQ_OPS      => [ 'eq_re' ],
   SRC1_TYPES  => [ @complex_types ],
   FUNCS       => [ 'dot' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("r = sum( I )");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   DEST_SCALAR => 1,
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'sum' ],
   SRC2_TYPES  => [ 'Int' ],
  )}});

comment1("<type> = sum( <field> )");

make_functions(%{{
  (
   DEST_TYPES  => [ @float_types ],
   DEST_SCALAR => 1,
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'sum' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment1("r[i] = norm2( <field>, subset[i] )");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   DEST_SCALAR => 1,
   DEST_MULTI  => 1,
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'norm2' ],
   SRC2_TYPES  => [ @shift_types ],
  )}});

comment1("r[i] = <field> dot <field> on subset[i]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   DEST_SCALAR => 1,
   DEST_MULTI  => 1,
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'Int','Real' ],
   FUNCS       => [ 'dot' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("c[i] = <field> dot <field> on subset[i]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Complex' ],
   DEST_SCALAR => 1,
   DEST_MULTI  => 1,
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ @complex_types ],
   FUNCS       => [ 'dot' ],
   SRC2_TYPES  => [ 'SRC1' ],
  )}});

comment1("r[i] = sum( I ) on subset[i]");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   DEST_SCALAR => 1,
   DEST_MULTI  => 1,
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'sum' ],
   SRC2_TYPES  => [ 'Int' ],
  )}});

comment1("<type>[i] = sum( <field> ) on subset[i]");

make_functions(%{{
  (
   DEST_TYPES  => [ @float_types ],
   DEST_SCALAR => 1,
   DEST_MULTI  => 1,
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'sum' ],
   SRC2_TYPES  => [ 'DEST' ],
  )}});

comment0("fills");
comment1("<field> = 0");

make_functions(%{{
  (
   DEST_TYPES  => [ @shift_types ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'zero' ],
  )}});

comment1("<field> = <QLA field>");

make_functions(%{{
  (
   DEST_TYPES  => [ @shift_types ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'DEST' ],
   SRC1_SCALAR => 1,
  )}});

comment1("M = c");

make_functions(%{{
  (
   DEST_TYPES  => [ 'ColorMatrix' ],
   EQ_OPS      => [ 'eq' ],
   SRC1_TYPES  => [ 'Complex' ],
   SRC1_SCALAR => 1,
  )}});

comment1("S = seed i I");

make_functions(%{{
  (
   DEST_TYPES  => [ 'RandomState' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'seed_i' ],
   EXTRA_ARGS2 => "QLA_Int i, ",
   SRC2_TYPES  => [ 'Int' ],
  )}});

comment1("R = uniform S");

make_functions(%{{
  (
   DEST_TYPES  => [ 'Real' ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'random' ],
   SRC2_TYPES  => [ 'RandomState' ],
  )}});

comment1("<field> = gaussian S");

make_functions(%{{
  (
   DEST_TYPES  => [ @float_types ],
   EQ_OPS      => [ 'eq' ],
   FUNCS       => [ 'gaussian' ],
   SRC2_TYPES  => [ 'RandomState' ],
  )}});
