#include <stdio.h>
//#define QDP_Precision 'F'
//#define QDP_Nc 3
#include <qdp.h>

void
initial_gauge(QLA_ColorMatrix *g, int coords[])
{
  QLA_Complex z;
  int i;

  QLA_M_eq_zero(g);
  for(i=0; i<QLA_Nc; i++) {
    QLA_c_eq_r_plus_ir(z, i, 0);
    QLA_elem_M(*g,i,i) = z;
  }
}

void
initial_df(QLA_DiracFermion *df, int coords[])
{
  QLA_Complex z;
  int i,j;

  for(i=0; i<QLA_Nc; i++) {
    for(j=0; j<4; j++) {
      QLA_c_eq_r_plus_ir(z, i+j, coords[0]);
      QLA_elem_D(*df,i,j) = z;
    }
  }
}

void
initial_li(QLA_Int *li, int coords[])
{
  int i,t;

  t = coords[3];
  for(i=2; i>=0; --i) {
    t = t*4 + coords[i];
  }
  *li = t;
}

void
print_df(QLA_DiracFermion *df, int coords[])
{
  QLA_Complex z;
  int i,j;

  for(i=0; i<4; i++) if(coords[i]) return;
  for(i=0; i<QLA_Nc; i++) {
    for(j=0; j<4; j++) {
      z = QLA_elem_D(*df,i,j);
      printf("%g\t%g\n", QLA_real(z), QLA_imag(z));
    }
  }
}

int
timeslices(int x[])
{
  return x[3];
}

int
main(int argc, char *argv[])
{
  int lattice_size[4] = { 4,4,4,8 };
  QDP_ColorMatrix *m;
  QDP_DiracFermion *d1, *d2, *d3;
  QDP_RandomState *rs;
  QDP_Int *li;
  QDP_Subset *ts;

  QDP_initialize(argc, argv);
  QDP_set_latsize(4, lattice_size);
  QDP_create_layout();

  m = QDP_create_M();
  d1 = QDP_create_D();
  d2 = QDP_create_D();
  d3 = QDP_create_D();
  rs = QDP_create_S();
  li = QDP_create_I();

  ts = QDP_create_subset(timeslices, lattice_size[3]);
  //fprintf(stderr,"test1\n");

  //QDP_M_eq_func(m, initial_gauge, QDP_all);
  //QDP_D_eq_func(d1, initial_df, QDP_all);

  QDP_I_eq_func(li, initial_li, QDP_all);
  //fprintf(stderr,"test2\n");
  QDP_S_eq_seed_i_I(rs, 987654321, li, QDP_all);
  //fprintf(stderr,"test3\n");

  QDP_M_eq_gaussian_S(m, rs, QDP_all);
  QDP_D_eq_gaussian_S(d1, rs, QDP_all);

  //QDP_F3_M_eq_zero(g, QDP_all);
  //QDP_F3_D_eq_zero(d1, QDP_all);
  //QDP_F3_D_eq_zero(d2, QDP_all);
  //QDP_F3_D_eq_zero(d3, QDP_all);

  //fprintf(stderr,"%i %i\n",((int *)QDP_shift_nn[0])[0],((int *)QDP_shift_nn[0])[1]);
  QDP_D_eq_sD(d2, d1, QDP_neighbor[0], 1, QDP_all);
  //QDP_D_eq_D(d2, d1, QDP_all);

  QDP_D_eq_M_times_D(d3, m, d2, QDP_all);

  QDP_D_eq_func(d3, print_df, QDP_all);
  printf("\n");
  QDP_D_eq_zero(d3, QDP_odd);
  QDP_D_eq_func(d3, print_df, QDP_all);
  printf("\n");
  //QDP_D_eq_zero(d3, QDP_even);
  //QDP_D_eq_func(d3, print_df, QDP_all);
  QDP_D_eq_zero(d3, ts[1]);
  QDP_D_eq_func(d3, print_df, QDP_all);
  printf("\n");
  QDP_D_eq_zero(d3, ts[0]);
  QDP_D_eq_func(d3, print_df, QDP_all);
  printf("\n\n");
  QDP_D_eq_zero(d2, QDP_odd);
  QDP_D_eq_func(d2, print_df, QDP_all);
  printf("\n");
  QDP_D_eq_zero(d2, QDP_even);
  QDP_D_eq_func(d2, print_df, QDP_all);
  printf("\n");
  QDP_D_eq_sD(d2, d1, QDP_neighbor[0], 1, QDP_odd);
  QDP_D_eq_func(d2, print_df, QDP_all);
  printf("\n");
  QDP_D_eq_sD(d2, d1, QDP_neighbor[0], 1, QDP_even);
  QDP_D_eq_func(d2, print_df, QDP_all);
  printf("\n");

  QDP_expose_D(d2);
  QDP_reset_D(d2);
  QDP_D_eq_zero(d2, QDP_all);

  QDP_destroy_M(m);
  QDP_destroy_D(d1);
  QDP_destroy_D(d2);
  QDP_destroy_D(d3);

  QDP_finalize();
  return 0;
}
